#!/usr/bin/env python
# -*- coding: utf-8 -*-
# __init__.py
"""Package initialization for dicompyler-core."""
# Copyright (c) 2009-2016 Aditya Panchal
# This file is part of dicompyler-core, relased under a BSD license.
#    See the file license.txt included with this distribution, also
#    available at https://github.com/dicompyler/dicompyler-core/

__author__ = 'Aditya Panchal'
__email__ = 'apanchal@bastula.org'
__version__ = '0.5.4'
__version_info__ = (0, 5, 0)
